from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import guess_lexer_for_filename, get_lexer_by_name
from pygments.util import ClassNotFound
from xml.etree import ElementTree
from . import Route, url
from ..model import Build, Project
import os


@url(r'/project/browse/(\d+)/path/(.+)', suffix='LastPath')
@url(r'/project/browse/(\d+)', suffix='Last')
@url(r'/project/browse/(\d+)/build/(\d+)/path/(.+)')
class ProjectBrowse(Route):
    def get_tree_dict(self, directory, path):
        tree_dict = {}
        base, prefix = os.path.split(directory)

        for dir, dirs, files in os.walk(directory):
            dir = dir.replace(base + '/', '')
            parts = dir.split(os.path.sep)
            dct = tree_dict
            for part in parts[1:]:
                dct = dct['dirs'][part]

            dct['name'] = (dir.replace(prefix + '/', '')
                           if dir != prefix else '')
            dct['active'] = path.startswith(dct['name'] + '/')
            dct['dirs'] = {dir_: {} for dir_ in dirs}
            dct['files'] = {file: os.path.join(dct['name'], file) == path
                            for file in files}

        return tree_dict

    def recurse(self, dct, id, build):
        out = '<ul>'
        for name, dirdct in sorted(dct['dirs'].items()):
            out += '<li class="dir">'
            out += '<input type="checkbox"%s id="node_%s" />' % (
                ' checked' if dirdct['active'] else '',
                dirdct['name'].replace('/', ':'))
            out += '<label class="dir_title" for="node_%s">' % (
                dirdct['name'].replace('/', ':'))
            out += '<i class="glyphicon"></i> %s' % name
            out += '</label>%s' % self.recurse(dirdct, id, build)
            out += '</li>'

        for file, active in sorted(dct['files'].items()):
            out += '<li class="file">'
            out += '<label>'
            if not active:
                out += '<a href="%s">' % self.reverse_url(
                    'ProjectBrowse', id, build.build_id,
                    os.path.join(dct['name'], file))
            out += '<i class="glyphicon glyphicon-file"></i>%s' % file
            if not active:
                out += '</a>'
            out += '</label>'
            out += '</li>'

        return out + '</ul>'

    def get(self, id, build_id=None, path=''):
        if path is None and build_id is not None:
            path = build_id
            build_id = None
        code = ''
        formatter = HtmlFormatter(linenos=True, cssclass='code')

        project = self.db.query(Project).get(id)
        if build_id:
            build = self.db.query(Build).get((build_id, project.project_id))
        else:
            build = project.last_build

        if path and '..' not in path:
            file = os.path.join(project.src_dir, path)
            if os.path.exists(file):
                try:
                    with open(file, 'r') as f:
                        code = f.read()
                except UnicodeDecodeError:
                    code = None
                else:
                    try:
                        lexer = guess_lexer_for_filename(file, code)
                    except ClassNotFound:
                        lexer = get_lexer_by_name('text')
                    code = highlight(code, lexer, formatter)

        self.render(
            'source.html', code=code,
            tree=self.recurse(
                self.get_tree_dict(project.src_dir, path), id, build),
            project=project,
            build_id=build.build_id,
            path=path
        )

    def get_coverage_formatter(self, id, build, file):
        project = self.db.query(Project).get(id)
        build = self.db.query(Build).get((build.build_id, project.project_id))
        coverage_file = os.path.join(
            build.dir, build.project.coverage.coverage_path)
        tree = ElementTree.parse(coverage_file)
        root = tree.getroot()
        packages = root.find('packages')
        coverage = []
        for package in packages:
            for classes in package:
                for cls in classes:
                    if file.endswith(cls.get('filename', '')):
                        for lines in cls:
                            for line in lines:
                                if line.get('hits', '0') > '0':
                                    coverage.append(line.get('number'))
        return HtmlFormatter(
            linenos=True, cssclass='code', hl_lines=coverage)


@url(r'/project/browse/raw/(\d+)/(\d+)/(.*)')
class BrowseRaw(Route):
    def get(self, id, build, path):
        project = self.db.query(Project).get(id)
        if path and '..' not in path:
            file = os.path.join(project.src_dir, path)
            if os.path.exists(file):
                with open(file, 'rb') as f:
                    self.write(f.read())
        self.finish()



# -*- coding: utf-8 -*-
#
# ymci - You Modern Continuous Integration server
# Copyright © 2014 Florian Mounier, Kozea

import os
import yaml


class Config(dict):

    def __init__(self, path):
        super().__init__()
        self.update({
            'projects_path': 'projects',
            'projects_realpath': os.path.realpath(
                os.path.join(
                    os.path.dirname(os.path.dirname(__file__)), 'projects')),
            'db_url': 'postgresql+psycopg2://ymci@localhost/ymci'
        })

        self._read(path)
        self._sync()

    def _read(self, path):
        if os.path.exists(path):
            with open(path, 'r') as f:
                self.update(yaml.load(f))
        self['path'] = path

    def _sync(self):
        with open(self['path'], 'w') as f:
            f.write(yaml.safe_dump(
                dict(self), default_flow_style=False, allow_unicode=True))



from . import ymci_style, graph_config
from .. import url, Route, server, WebSocket
from ..model import Project, Build
from ..utils import short_transaction, secure_rmtree
from datetime import datetime
from logging import getLogger
from wtforms_alchemy import ModelForm, ModelFormField
import pygal
import os


log = getLogger('ymci')


class Meta(object):
    model = Project

forms = {
    c.Meta.model.project.property.back_populates: ModelFormField(c)
    for c in server.plugins['ymci.ext.form.Form']}

for ufield in forms.values():
    ufield.creation_counter += 10000

forms.update({'Meta': Meta})
ProjectForm = type('ProjectForm', (ModelForm,), forms)


@url(r'/project/list')
class ProjectList(Route):
    def get(self):
        self.render(
            'project/list.html', projects=self.db.query(Project).all())


@url(r'/project/view/(?P<project_id>\d+)')
class ProjectView(Route):
    def get(self, project_id):
        project = self.db.query(Project).get(project_id)
        self.render('project/view.html', project=project)


@url(r'/project/add')
class ProjectAdd(Route):
    def get(self):
        return self.render(
            'form.html', form=ProjectForm(),
            render_form_recursively=self.render_form_recursively)

    def post(self):
        form = ProjectForm(self.posted)
        if form.validate():
            project = Project()
            for hook in self.application.plugins['ymci.ext.hook.FormHook']:
                hook().pre_populate(form)
            form.populate_obj(project)
            self.db.add(project)
            self.db.commit()
            self.blocks.project.refresh()
            self.blocks.home.refresh()
            self.redirect('/')
            return
        self.render(
            'form.html', form=form,
            render_form_recursively=self.render_form_recursively)


@url(r'/project/edit/(?P<project_id>\d+)')
class ProjectEdit(Route):
    def get(self, project_id):
        project = self.db.query(Project).get(project_id)
        self.render(
            'form.html', form=ProjectForm(obj=project),
            render_form_recursively=self.render_form_recursively)

    def post(self, project_id):
        project = self.db.query(Project).get(project_id)
        form = ProjectForm(self.posted, project)
        if form.validate():
            for hook in self.application.plugins['ymci.ext.hook.FormHook']:
                hook().pre_populate(form)
            old_dir = project.project_dir
            form.populate_obj(project)
            if project.project_dir != old_dir:
                os.rename(old_dir, project.project_dir)
            self.db.commit()
            self.blocks.project.refresh()
            self.blocks.home.refresh()
            self.redirect('/')
            return
        self.render(
            'form.html', form=form,
            render_form_recursively=self.render_form_recursively)


@url(r'/project/delete/(?P<project_id>\d+)')
class ProjectDelete(Route):
    def get(self, project_id):
        project = self.db.query(Project).get(project_id)
        self.render(
            'ask.html', confirm="Do you really want to delete project %s" %
            project.name)

    def post(self, project_id):
        project = self.db.query(Project).get(project_id)
        self.db.delete(project)
        secure_rmtree(
            os.path.join(server.conf['projects_realpath'], project.dir_name))
        self.db.commit()
        self.blocks.project.refresh()
        self.blocks.home.refresh()
        self.redirect(self.reverse_url('ProjectList'))


@url(r'/project/build/(?P<project_id>\d+)')
class ProjectBuild(Route):
    def get(self, project_id):
        project = self.db.query(Project).get(project_id)
        build = Build()
        build.timestamp = datetime.now()
        build.build_id = (
            project.last_build and project.last_build.build_id or 0) + 1
        build.status = 'PENDING'
        project.builds.append(build)

        self.db.add(build)

        self.db.commit()

        self.application.builder.add(build)

        self.redirect(self.reverse_url(
            'ProjectLog', project.project_id, build.build_id))


@url(r'/log/(\d+)/(\d+)/pipe')
class ProjectLogWebSocket(WebSocket):
    def open(self, id, idx):
        self.id = id
        self.idx = idx
        with short_transaction() as db:
            self.build = db.query(Build).get((idx, id))
            self.application.builder.log_streams['%s-%s' % (
                self.build.project_id,
                self.build.build_id
            )].append(self)

            if os.path.exists(self.build.log_file):
                with open(self.build.log_file, 'r') as f:
                    for line in f:
                        self.write_message(line)

    def on_close(self):
        self.application.builder.log_streams[
            '%s-%s' % (self.id, self.idx)].remove(self)


@url(r'/project/log/(?P<project_id>\d+)/(?P<idx>\d+)')
@url(r'/project/log/(?P<project_id>\d+)', suffix='Last')
class ProjectLog(Route):
    def get(self, project_id, idx=None):
        project = self.db.query(Project).get(project_id)
        if idx:
            build = self.db.query(Build).get((idx, project_id))
        else:
            build = project.last_build
        if build.status != 'RUNNING' and os.path.exists(build.log_file):
            with open(build.log_file, 'r') as f:
                log = f.read()
        else:
            log = None
        self.render(
            'project/log.html', project=project, build=build, log=log)


@url(r'/project/build/(?P<project_id>\d+)/(?P<idx>\d+)/stop')
class ProjectBuildStop(Route):
    def get(self, project_id, idx):
        build = self.db.query(Build).get((idx, project_id))
        self.application.builder.stop(build)

        self.redirect(self.reverse_url(
            'ProjectLog', build.project_id, build.build_id))


@url(r'/project/chart/time/(?P<project_id>\d+).svg')
@url(r'/project/chart/time/'
     '(?P<project_id>\d+)_(?P<width>\d+)_(?P<height>\d+).svg',
     suffix='Size')
class ProjectChartTime(Route):
    def get(self, project_id, width=None, height=None):
        project = self.db.query(Project).get(project_id)
        config = graph_config(width, height)
        config.style = ymci_style
        svg = pygal.Line(config)
        builds = project.builds.filter(Build.status == 'SUCCESS').all()[::-1]
        svg.add('Success', [{
            'xlink': self.reverse_url('ProjectLog', project_id, b.build_id),
            'value': b.duration} for b in builds])
        if width and height:
            svg.x_labels = ['#%d' % b.build_id for b in builds]
        svg.value_formatter = lambda x: '%.2fs' % (x or 0)
        svg.interpolate = 'cubic'
        svg.show_legend = False
        svg.title = 'Build duration in seconds'
        self.set_header("Content-Type", "image/svg+xml")
        self.write(svg.render())



from alembic import command
from alembic.config import Config


def upgrade():
    alembic_cfg = Config("alembic.ini")
    command.revision(alembic_cfg, "Automatic revision", autogenerate=True)
    command.upgrade(alembic_cfg, "head")



from ymci.utils import pretty_duration, pretty_duration_round


def test_pretty_duration():
    assert pretty_duration(0) == ''
    assert pretty_duration(.001) == '1ms'
    assert pretty_duration(.125178213) == '125ms'
    assert pretty_duration(1) == '1s'
    assert pretty_duration(1.12) == '1s 120ms'
    assert pretty_duration(45) == '45s'
    assert pretty_duration(1000) == '16min 40s'
    assert pretty_duration(60) == '1min'
    assert pretty_duration(60.01) == '1min 10ms'
    assert pretty_duration(111111111.111) == (
        '3 years 6 months 11 days 11min 51s 111ms')


def test_pretty_duration_round():
    assert pretty_duration_round(0) == ''
    assert pretty_duration_round(.001) == '1ms'
    assert pretty_duration_round(.125178213) == '125ms'
    assert pretty_duration_round(1) == '1s'
    assert pretty_duration_round(1.12) == '1s'
    assert pretty_duration_round(45) == '1min'
    assert pretty_duration_round(1000) == '17min'
    assert pretty_duration_round(60) == '1min'
    assert pretty_duration_round(60.01) == '1min'
    assert pretty_duration_round(111111111.111) == '4 years'



# -*- coding: utf-8 -*-
#
# ymci - You Modern Continuous Integration server
# Copyright © 2014 Florian Mounier, Kozea

from tornado.web import (
    Application, RequestHandler, url as tornado_url, HTTPError)
from sqlalchemy.orm import scoped_session, sessionmaker
from tornado.escape import json_decode, to_unicode
from tornado.websocket import WebSocketHandler
from tornado.options import define, parse_command_line, options
from tornado.ioloop import IOLoop
from collections import defaultdict
from tornado.web import StaticFileHandler
from logging import getLogger
from .config import Config
import os.path

__version__ = '0.0.5'

log = getLogger('ymci')


define("debug", default=False, help="Debug mode")
define("host", default='ymci.l', help="Server host")
define("port", default=7361, help="Server port")
define("config", default='ymci.yaml', help="YMCI config file")
define("secret", default='secret', help="Secret key for cookies")
define("upgrade", default=False, help="Upgrade DB")

parse_command_line()
ioloop = IOLoop.instance()

MESSAGE_LEVELS = ['success', 'info', 'warning', 'danger']


class ExtStaticFileHandler(StaticFileHandler):
    ext_path = {}

    @classmethod
    def get_absolute_path(cls, root, path):
        if path.startswith('ext'):
            key = '/'.join(path.split('/')[:2])
            path = '/'.join(path.split('/')[2:])
            root = ExtStaticFileHandler.ext_path[key]
        abspath = os.path.abspath(os.path.join(root, path))
        return abspath

    def validate_absolute_path(self, root, absolute_path):
        if self.path.startswith('ext'):
            key = '/'.join(self.path.split('/')[:2])
            root = ExtStaticFileHandler.ext_path[key]
        return super().validate_absolute_path(root, absolute_path)

server = Application(
    debug=options.debug,
    cookie_secret=options.secret,
    static_path=os.path.join(os.path.dirname(__file__), "static"),
    static_handler_class=ExtStaticFileHandler,
    template_path=os.path.join(os.path.dirname(__file__), "templates"),
    login_url="/auth/login")

server.conf = Config(options.config)

from .model import engine, Query
server.scoped_session = scoped_session(
    sessionmaker(bind=engine, query_cls=Query))


class url(object):
    def __init__(self, url, name=None, suffix=None):
        self.url = url
        self.name = name
        self.suffix = suffix or ''

    def __call__(self, cls):
        server.add_handlers(
            r'.*$',
            (tornado_url(self.url, cls, name=self.name or (
                cls.__name__ + self.suffix)),)
        )
        return cls


class MultiDict(dict):
    def getlist(self, attr):
        return [v.decode('utf-8').replace('\r', '')
                if isinstance(v, bytes) else v
                for v in (self[attr] if isinstance(self[attr], list)
                else [self[attr]])]


class Base(object):
    def get_template_namespace(self):
        from . import utils
        namespace = super().get_template_namespace()
        namespace.update(dict(
            server=self.application,
            utils=utils
        ))
        return namespace

    @property
    def log(self):
        return log

    @property
    def blocks(self):
        return server.components.blocks


class Route(Base, RequestHandler):
    @property
    def posted(self):
        return MultiDict(self.request.arguments)

    def abort(self, code):
        raise HTTPError(code)

    @property
    def db(self):
        # Explicit creation of the request session
        if not getattr(self, '_db', None):
            # Cache it even if it's the same
            self._db = self.application.scoped_session()
        return self._db

    def get_current_user(self):
        user = self.get_secure_cookie("user")
        if not user:
            return None
        return json_decode(to_unicode(user))

    def set_flash_message(self, key, message):
        if key not in MESSAGE_LEVELS:
            log.error("This flash message will not appear because the key '%s'"
                      " doesn't match any class of bootstrap" % key)
        self.set_secure_cookie('flash_message_%s' % key, message)

    def get_flash_messages(self):
        messages = {}
        for level in MESSAGE_LEVELS:
            message = self.get_secure_cookie('flash_message_%s' % level)
            messages.update({'%s' % level: message})
            self.clear_cookie('flash_message_%s' % level)
        return messages

    def on_finish(self):
        # Teardown of the current session
        return self.application.scoped_session.remove()

    def render_form_recursively(self, form):
        return self.render_string(
            'fields.html', form=form,
            render_form_recursively=self.render_form_recursively)

    def prepare(self):
        for Hook in self.application.plugins['ymci.ext.hooks.PrepareHook']:
            Hook(self.db).prepare(self)


class WebSocket(Base, WebSocketHandler):
    pass


class Container(object):
    def __iter__(self):
        for name, attr in self.__dict__.items():
            if not name.startswith('_'):
                yield attr


class Components(object):
    def __init__(self):
        self.blocks = Container()
        self.project_charts = Container()
        self.project_links = Container()

server.components = Components()


class MetaBlock(type):
    def __new__(mcs, *args, **kwargs):
        cls = super().__new__(mcs, *args, **kwargs)
        if cls.__name__ != 'BlockWebSocket':
            cls._sockets = defaultdict(set)
        return cls


class BlockWebSocket(WebSocket, metaclass=MetaBlock):
    def open(self, *args):
        self.args = args
        self.__class__._sockets[self.args].add(self)
        self.render()

    @classmethod
    def refresh(cls, *args):
        args = tuple(map(str, args))
        for sock in cls._sockets[args]:
            sock.render()

    def render_block(self):
        pass

    def write_message(self, message):
        super().write_message(message)

    def render(self):
        # Scope the session in the request
        self.db = self.application.scoped_session()
        self.write_message(self.render_block(*self.args))
        self.application.scoped_session.remove()

    def on_close(self):
        self.__class__._sockets[self.args].remove(self)

from .builder import Pool
server.builder = Pool()

from .ext import Plugins
server.plugins = Plugins()
server.plugins.init()


import ymci.routes



from .. import url, BlockWebSocket, server
from ..model import Project, Build
from time import time


@url(r'/blocks/build')
class BuildBlock(BlockWebSocket):
    def render_block(self):
        builder = server.builder
        builds = []
        for task in builder.tasks:
            build = self.db.query(Build).get(
                (task.build_id,
                 task.project_id))
            if task.start_time:
                now = time() - task.start_time
            else:
                now = None
            builds.append((now, build))

        free_slots = builder.limit - len(builder.tasks)
        return self.render_string(
            'blocks/build.html', current_builds=builds, free_slots=free_slots)


@url(r'/blocks/project')
class ProjectBlock(BlockWebSocket):
    def render_block(self):
        return self.render_string(
            'blocks/project.html',
            projects=self.db.query(Project).all())


@url(r'/blocks/history/([^\/]+)')
class HistoryBlock(BlockWebSocket):
    def render_block(self, id):
        return self.render_string(
            'blocks/history.html',
            project=self.db.query(Project).get(id))


@url(r'/blocks/home')
class HomeBlock(BlockWebSocket):
    def render_block(self):
        return self.render_string(
            'blocks/home.html',
            projects=self.db.query(Project).all())


@url(r'/blocks/plugins')
class PluginBlock(BlockWebSocket):
    def render_block(self):
        plugins = server.plugins['ymci.ext']
        return self.render_string('blocks/plugins.html', plugins=plugins)


blocks = server.components.blocks
blocks.build = BuildBlock
blocks.project = ProjectBlock
blocks.home = HomeBlock
blocks.history = HistoryBlock
blocks.plugins = PluginBlock



from .. import url, Route
from ..model import Project
from logging import getLogger
import pygal
from pygal import Config
from copy import copy
log = getLogger('ymci')


base_style = pygal.style.Style(
    background='white',
    plot_background='white',
    foreground='#555',
    foreground_light='#666',
    foreground_dark='#444',
    opacity='.6',
    opacity_hover='.9',
    transition='400ms ease-in',
    colors=[])

ymci_style = pygal.style.RotateStyle('#28b62c', base_style=base_style)


def graph_config(width, height):
    class GraphConfig(Config):
        """Config for minimal graph."""
        js = ['/static/svg.jquery.js?://', '/static/pygal-tooltips.js?://']
        style = pygal.style.Style(**ymci_style.__dict__)
        show_minor_x_labels = False
        x_labels_major_count = 20
        truncate_label = 10
        legend_at_bottom = True
    config = GraphConfig()
    if not (width or height):
        config.show_legend = False
        config.show_y_labels = False
        config.show_x_labels = False
        config.show_dots = False
        config.no_data_font_size = 36
        config.x_labels_major_count = None
        config.width = config.height = 200
    else:
        # Good ratio for graph size, associated with a `margin` on the popup
        config.width = int(width) * 0.8
        config.height = int(height) * 0.8
    return copy(config)


@url(r'/')
class Index(Route):
    def get(self):
        return self.render(
            'index.html', projects=self.db.query(Project).all())

import ymci.routes.blocks
import ymci.routes.project
import ymci.routes.browse
import ymci.routes.auth
import ymci.routes.admin



from .. import url, Route
from ..model import User
from wtforms.form import Form
from wtforms.fields import StringField, PasswordField
from wtforms.validators import InputRequired
from tornado.escape import json_encode
from simplepam import authenticate
import datetime


class AuthForm(Form):
    form_name = 'Login'
    login = StringField('login', validators=[InputRequired])
    password = PasswordField('password', validators=[InputRequired])


@url(r'/auth/login')
class AuthLogin(Route):
    def get(self):
        self.render('form.html', form=AuthForm(),
                    render_form_recursively=self.render_form_recursively)

    def post(self):
        form = AuthForm(self.posted)
        service = self.read_pam_config()
        check = authenticate(form.login.data, form.password.data, service)
        if check:
            user = (self.db.query(User)
                    .filter(User.login == form.login.data)
                    .first())

            if not user:
                user = User(login=form.login.data)
                self.db.add(user)
                self.db.commit()

            self.set_current_user(user.login)
            self.set_flash_message('success', 'Connection successful')
            user.login_count = (user.login_count or 0) + 1
            user.last_login = datetime.datetime.now()
            self.db.commit()
            return self.redirect(self.get_argument("next", "/"))

        self.set_flash_message('danger', 'User/Password do not match')
        self.redirect(self.reverse_url('AuthLogin'))

    def set_current_user(self, user):
        if user:
            self.set_secure_cookie("user", json_encode(user))
        else:
            self.clear_cookie("user")

    def read_pam_config(self):
        return self.application.conf.get('pam_service', 'login')


@url(r'/auth/logout')
class AuthLogout(Route):
    def get(self):
        self.clear_cookie("user")
        self.set_flash_message('success', 'Logout successful')
        self.redirect("/")



import sys
import site
import pkg_resources
from logging import getLogger
from importlib import reload, invalidate_caches

log = getLogger('ymci')


class Plugins(dict):
    def __init__(self):
        super().__init__()

    def __getitem__(self, entry):
        if entry not in self:
            self.load(entry)
        return super().__getitem__(entry)

    def load(self, key):
        self[key] = []
        for entry in pkg_resources.iter_entry_points(key):
            try:
                cls = entry.load()
                cls.__entry_name__ = entry.name
                self[key].append(cls)
                log.info(
                    'Plugin %r for %s successfully loaded' % (entry, key))
            except Exception:
                log.exception('Failed to load plugin %r for %s' % (entry, key))

    def init(self):
        self.load('ymci.ext')

        # Explicit loads
        self.load('ymci.ext.db.Table')
        self.load('ymci.ext.routes.Route')

    def reload(self):
        self.clear()

        # ?
        invalidate_caches()

        # Clean up sys.path
        sys.path[:] = list(filter(
            lambda x: x.startswith(sys.prefix),
            sys.path))

        # Reloading site for egg packages
        reload(site)

        # Fix basestring import bug
        del pkg_resources.basestring

        # Reload packgares
        reload(pkg_resources)

        self.init()



from tornado.web import HTTPError
from sqlalchemy import (
    Column, Integer, String, Text, Float, DateTime, ForeignKey, create_engine,
    func)
from sqlalchemy.orm import relationship, Query as SAQuery
from sqlalchemy.ext.declarative import declarative_base
from logging import getLogger
from .. import server
import os
import datetime


log = getLogger('ymci')
engine = create_engine(server.conf['db_url'], echo=False)
Table = declarative_base()


class Query(SAQuery):
    def get(self, *args, **kwargs):
        obj = super(Query, self).get(*args, **kwargs)
        if obj is None:
            raise HTTPError(404)
        return obj


class Project(Table):
    __tablename__ = 'project'
    project_id = Column(Integer, autoincrement=True, primary_key=True)
    name = Column(String, nullable=False)
    description = Column(Text)
    repository = Column(String)
    script = Column(Text)

    builds = relationship('Build', cascade='all, delete-orphan',
                          backref='project', lazy='dynamic',
                          order_by='Build.build_id.desc()')

    @property
    def dir_name(self):
        return (
            self.name
            .replace(' ', '_')
            .replace('.', '_')
            .replace('\\', '_')
            .replace('/', '_'))

    @property
    def project_dir(self):
        path = os.path.join(server.conf['projects_path'], self.dir_name)
        if not os.path.exists(path):
            os.mkdir(path)
        return path

    @property
    def src_dir(self):
        path = os.path.join(self.project_dir, 'src')
        if not os.path.exists(path):
            os.mkdir(path)
        return path

    @property
    def last_build(self):
        return self.builds.order_by(Build.build_id.desc()).first()

    @property
    def last_non_running_build(self):
        return (
            self.builds
                .filter(~Build.status.in_(('RUNNING', 'PENDING')))
                .order_by(Build.build_id.desc())
                .first())

    @property
    def last_successful_build(self):
        return (
            self.builds
                .filter(Build.status == 'SUCCESS')
                .order_by(Build.build_id.desc())
                .first())

    @property
    def average_build_duration(self):
        return (self.builds
                .filter(Build.status == 'SUCCESS')
                .order_by(Build.build_id.desc())
                .limit(5)
                .from_self(
                    func.avg(Build.duration)).scalar() or 60 * 10)

    def health(self, over=10):
        count = self.builds.filter(
            Build.status.in_(('SUCCESS', 'FAILED'))).count()

        if not count:
            return 0

        if count > over:
            count = over

        ok = len(list(filter(
            lambda b: b.status == 'SUCCESS', self.builds.filter(
                Build.status.in_(('SUCCESS', 'FAILED')))[:count])))
        return int((ok / count) * over)


class Build(Table):
    __tablename__ = 'build'
    build_id = Column(Integer, primary_key=True)
    project_id = Column(Integer, ForeignKey(Project.project_id),
                        primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    duration = Column(Float)
    status = Column(String)

    @property
    def dir(self):
        return os.path.join(
            self.project.project_dir, 'build_%d' % self.build_id)

    @property
    def log_file(self):
        return os.path.join(
            self.project.project_dir,
            'build_%d.log' % self.build_id)

    @property
    def bootstrap_status(self):
        return {
            'SUCCESS': 'success',
            'RUNNING': 'primary',
            'BROKEN': 'danger',
            'FAILED': 'warning',
            'STOPPED': 'default',
            'PENDING': 'info',
            None: 'default'
        }[self.status]


class User(Table):
    __tablename__ = 'ymci_user'

    login = Column(String, primary_key=True)
    last_login = Column(DateTime, default=datetime.datetime.now())
    login_count = Column(Integer)



from .. import url, Route
from ..utils.db import upgrade


@url(r'/admin')
class Admin(Route):
    def get(self):
        self.render('admin/index.html')

    def post(self):
        upgrade()
        self.redirect(self.reverse_url('Admin'))


@url(r'/admin/db_upgrade')
class AdminDBUpgrade(Route):
    def get(self):
        upgrade()
        self.redirect(self.reverse_url('Admin'))


@url(r'/admin/plugin_refresh')
class AdminPluginRefresh(Route):
    def get(self):
        self.application.plugins.reload()
        self.application.components.blocks.plugins.refresh()
        self.redirect(self.reverse_url('Admin'))



from ..builder.util import execute


class BuildHook(object):

    def __init__(self, build, out):
        self.build = build
        self.out = out

    def execute(self, cmd, cwd=None):
        def send(data):
            if data:
                self.out(data)

        execute(cmd, cwd or self.build.project.src_dir,
                send, self.build.project_id, self.build.build_id)

    @property
    def active(self):
        return False

    def pre_copy(self):
        pass

    def pre_build(self):
        pass

    def post_build(self):
        pass

    def validate_build(self):
        pass


class FormHook(object):
    @property
    def active(self):
        return False

    def pre_populate(self, form=None):
        pass

    def pre_commit(self, form=None):
        pass

    def pre_add(self):
        pass


class PrepareHook(object):
    def __init__(self, db):
        self.db = db

    @property
    def active(self):
        return False

    def prepare(self, route):
        pass



import sys
import os
from .. import (
    url, BlockWebSocket as BaseBlockWebSocket, Route as BaseRoute,
    ExtStaticFileHandler, MetaBlock as BaseMetaBlock)
from tornado.template import Template, BaseLoader, Loader
from logging import getLogger

log = getLogger('ymci')


class DualLoader(Loader):
    def __init__(self, roots, **kw):
        super().__init__('/', **kw)
        self.roots = roots
        self.root = None

    def load(self, name, parent_path=None):
        """Loads a template."""
        roots = self.roots
        if name.startswith('ymci:'):
            roots = roots[-1:]
            name = name[5:]
            parent_path = None
        for root in roots:
            try:
                self.root = root
                return super().load(name, parent_path)
            except FileNotFoundError:
                pass
            finally:
                self.root = self.roots[-1]
        raise FileNotFoundError(name, parent_path, self.roots)


class ExtPluginClass(type):
    def __init__(cls,  *args, **kwargs):
        super().__init__(*args, **kwargs)
        root_module = sys.modules.get(cls.__module__.split('.')[0])
        if hasattr(root_module, '__path__'):
            if hasattr(root_module.__path__, '_path'):
                cls._ext_dir = root_module.__path__._path[0]
            else:
                cls._ext_dir = root_module.__path__[0]
        else:
            cls._ext_dir = os.path.dirname(root_module.__file__)

        cls._key = os.path.join('ext', os.path.split(cls._ext_dir)[1])
        ExtStaticFileHandler.ext_path[cls._key] = os.path.join(
            cls._ext_dir, 'static')


class PathHandler(object):
    def get_template_path(self):
        # Look for plugin template path
        path = os.path.join(self._ext_dir, 'templates')
        if os.path.exists(path):
            return path

    def get_template_namespace(self):
        namespace = super().get_template_namespace()
        namespace.update(dict(
            static_url=self.local_static_url
        ))
        return namespace

    def local_static_url(self, file, *args, **kwargs):
        # Try to get from local plugin
        path = os.path.join(self._ext_dir, 'static', file)
        if os.path.exists(path):
            # Append ext prefix
            file = os.path.join(self._key, file)
        return self.static_url(file, *args, **kwargs)

    def create_template_loader(self, template_path):
        return DualLoader([
            template_path, self.application.settings.get("template_path")])


class Route(PathHandler, BaseRoute, metaclass=ExtPluginClass):
    pass


class ExtBlockClass(BaseMetaBlock, ExtPluginClass):
    pass


class BlockWebSocket(PathHandler, BaseBlockWebSocket, metaclass=ExtBlockClass):
    pass



from .. import server
import os
import shutil

periods = [
    (' year', 1000 * 60 * 60 * 24 * 365),
    (' month', 1000 * 60 * 60 * 24 * 30),
    (' day', 1000 * 60 * 60 * 24),
    ('h', 1000 * 60 * 60),
    ('min', 1000 * 60),
    ('s', 1000),
    ('ms', 1)
]


def pretty_duration(seconds):
    """Transform a float time in secoands to human readable duration."""
    ms = int((seconds or 0) * 1000)
    out = ''
    for name, nms in periods:
        if ms >= nms:
            val, ms = divmod(ms, nms)
            part = str(val) + name
            out += part
            if name in (' year', ' month', ' day') and val > 1:
                out += 's'
            out += ' '
    return out.strip()


def pretty_duration_round(seconds):
    """Transform a float time in secoands to human readable duration."""
    ms = int((seconds or 0) * 1000)
    out = ''
    for name, nms in periods:
        val = round(ms / nms)
        if val >= 1:
            part = str(val) + name
            out = part
            if name in (' year', ' month', ' day') and ms > 1:
                out += 's'
            return out
    return out


class short_transaction(object):
    def __enter__(self):
        self.db = server.scoped_session()
        return self.db

    def __exit__(self, type, value, traceback):
        if type is None:
            self.db.commit()
        server.scoped_session.remove()


def secure_rmtree(path):
    """Checks a path before deleting it."""
    path = os.path.realpath(path)
    if (os.path.exists(path) and os.path.isdir(path) and
            server.conf['projects_realpath'] in path):
        shutil.rmtree(path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ymci - Your Modern Continuous Integration server
# Copyright © 2014 Florian Mounier, Kozea

import os
import re

from setuptools import setup, find_packages
ROOT = os.path.dirname(__file__)

with open(os.path.join(ROOT, 'ymci', '__init__.py'), encoding='utf-8') as fd:
    __version__ = re.search("__version__ = '([^']+)'", fd.read()).group(1)

setup(
    name="ymci",
    version=__version__,
    description="Your Modern Continuous Integration server",
    author="Florian Mounier, Kozea",
    author_email="florian.mounier@kozea.fr",
    platforms="Linux",
    scripts=['ymci.server.py'],
    packages=find_packages(),
    provides=['ymci'],
    install_requires=['tornado', 'tornado_systemd', 'pyyaml', 'wtforms',
                      'sqlalchemy', 'psycopg2', 'wtforms_alchemy', 'pygal',
                      'pbkdf2', 'pygments', 'simplepam'],
    tests_require=["pytest"],
    package_data={'ymci': [
        'static/*.*',
        'static/*/*.*',
        'templates/*.*',
        'templates/*/*.*']},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Build Tools"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# ymci - You Modern Continuous Integration server
# Copyright © 2014 Florian Mounier, Kozea


from tornado.options import options
from tornado_systemd import SystemdHTTPServer
from ymci import server, ioloop
from ymci.utils.db import upgrade
from logging import getLogger
import sys

log = getLogger('ymci')
log.setLevel(10 if options.debug else 20)

if options.upgrade:
    upgrade()
    sys.exit(0)


if options.debug:
    from wdb.ext import wdb_tornado
    wdb_tornado(server, start_disabled=True)

http_server = SystemdHTTPServer(server)

log.debug('Listening')
http_server.listen(options.port)

ioloop.start()



from tornado.ioloop import IOLoop
from logging import getLogger
from collections import defaultdict
from threading import Thread
from time import time
from datetime import datetime
from .util import execute, ExecutionException
from ..utils import short_transaction
from ..model import Build
from .. import server
import traceback
import shutil
import os


log = getLogger('ymci')
blocks = server.components.blocks


def refresh(id):
    log.debug('Refreshing build/project/home/history')
    blocks.build.refresh()
    blocks.project.refresh()
    blocks.home.refresh()
    blocks.history.refresh(id)


class Pool(object):

    def __init__(self, ioloop=None):
        self.log_streams = defaultdict(list)
        self.queue = []
        self.tasks = []
        self.limit = server.conf.get('slot_size', 5)
        self.stop_flags = []
        self.ioloop = ioloop or IOLoop.instance()

    def add(self, build):
        task = Task(
            build.build_id, build.project_id,
            self.log_streams['%s-%s' % (build.project_id, build.build_id)],
            build.project.name, self.ioloop)
        if not len(self.queue) and len(self.tasks) < self.limit:
            self.build(task)
        else:
            self.queue.append(task)
        refresh(build.project_id)

    def stop(self, build):
        for task in self.tasks:
            if (task.project_id == build.project_id and
                    task.build_id == build.build_id):
                self.stop_flags.append((build.project_id, build.build_id))
                # Set stop flag to keep the fact that it's manual stopping
                task.stop = True

        for task in self.queue:
            if (task.project_id == build.project_id and
                    task.build_id == build.build_id):
                self.queue.remove(task)

        refresh(build.project_id)

    def build(self, task):
        self.tasks.append(task)
        task.callback = self.next
        self.ioloop.add_callback(task.start)

    def next(self, task):
        project_id = task.project_id
        self.tasks.remove(task)
        if len(self.queue):
            self.build(self.queue.pop(0))
        refresh(project_id)


class Task(Thread):
    def __init__(self, build_id, project_id, streams, name, ioloop,
                 *args, **kwargs):
        super(Task, self).__init__(*args, **kwargs)
        self.build_id = build_id
        self.project_id = project_id
        self.name = name
        self.streams = streams
        self.ioloop = ioloop
        self.stop = False
        self.start_time = None

    def out(self, data):
        self.log.write(data)
        self.log.flush()

        for stream in self.streams:
            try:
                for line in data.splitlines(True):
                    stream.write_message(line)
            except Exception:
                stream.on_close()
                stream.close()

    def read(self, data):
        if data:
            self.out(data)

    def run(self):
        try:
            self.safe_run()
        except Exception as e:
            try:
                log.exception('Error during task run')
                self.out('YMCI Internal error %s' % (
                    ', '.join(traceback.format_exception_only(type(e), e))))
            except Exception:
                pass
            with short_transaction() as db:
                build = db.query(Build).get((self.build_id, self.project_id))
                build.status = 'STOPPED'
            self.ioloop.add_callback(self.callback, self)

    def safe_run(self):
        log.info('Starting run for task %r on %s' % (self, datetime.now()))

        self.start_time = time()

        with short_transaction() as db:
            build = db.query(Build).get((self.build_id, self.project_id))
            build.status = 'RUNNING'

        self.db = server.scoped_session()
        self.build = self.db.query(Build).get((self.build_id, self.project_id))
        self.ioloop.add_callback(refresh, self.project_id)

        self.log = open(self.build.log_file, 'w')
        self.script = self.build.project.script

        def treat(e):
            if e.errno < 0:
                if self.stop:
                    self.out('\nStoped by user')
                    status = 'STOPPED'
                else:
                    self.out('\nCommand %s was killed by signal %d' % (
                        e.command, -e.errno))
                    status = 'KILLED'
            else:
                self.out('\nCommand %s returned %d' % (
                    e.command, -e.errno))
                status = 'BROKEN'
            return status

        log.debug('Running task %r' % self)
        try:
            self.run_task()
        except ExecutionException as e:
            log.warn('Error during task execution %r' % self, exc_info=True)
            self.build.status = treat(e)

        self.out('\n')
        log.debug('Task %r ran. Validating build' % self)
        try:
            for hook in self.build_hooks:
                hook.validate_build()
        except ExecutionException as e:
            log.warn('Error during task execution %r' % self, exc_info=True)
            self.build.status = treat(e)

        log.debug('Checking status after validation for %r ' % self)
        if self.build.status == 'RUNNING':
            self.out('Success !')
            self.build.status = 'SUCCESS'

        self.build.duration = time() - self.start_time
        self.out('\n(Duration %fs)' % self.build.duration)
        log.debug('Persisting state for %r ' % self)
        self.db.commit()

        log.debug('Running post build for %r ' % self)
        for hook in self.build_hooks:
            hook.post_build()

        log.debug('Closing log file and calling callback %r ' % self)
        self.log.close()
        self.ioloop.add_callback(self.callback, self)
        server.scoped_session.remove()

    def run_task(self):
        self.out('Starting build %d...\n' % self.build.build_id)
        self.build_hooks = []

        for Hook in server.plugins['ymci.ext.hooks.BuildHook']:
            def get_out(hook_name):
                def out(message):
                    self.out('%s> %s\n' % (hook_name, message))
                return out

            hook = Hook(self.build, get_out(Hook.__name__))
            if hook.active:
                self.build_hooks.append(hook)

        src = self.build.project.src_dir
        log.debug('Running pre copy hooks for %r' % self)
        for hook in self.build_hooks:
            hook.pre_copy()

        assert not os.path.exists(self.build.dir)

        log.debug('Copying tree for %r' % self)
        shutil.copytree(src, self.build.dir, symlinks=True)

        log.debug('Running pre build hooks for %r' % self)
        for hook in self.build_hooks:
            hook.pre_build()

        script_fn = os.path.abspath(os.path.join(self.build.dir, '.ymci.sh'))
        script = self.script
        if not script.startswith('#!'):
            # This is not sexy
            script = (
                '#!/bin/sh\n' +
                script
            )

        with open(script_fn, 'w') as build_script:
            build_script.write(script)
        os.chmod(script_fn, 0o700)

        log.debug('Executing %s for %r' % (script_fn, self))
        execute(
            script_fn,
            self.build.dir, self.read,
            self.build.project_id, self.build.build_id)
        log.debug('Execution complete for %r' % self)



import os
from subprocess import STDOUT
from tornado.process import Subprocess
from time import sleep
from signal import SIGKILL
from logging import getLogger
log = getLogger('ymci')


class ExecutionException(Exception):
    def __init__(self, command, errno):
        self.command = command
        self.errno = errno
        super(ExecutionException, self).__init__(
            '%s exited with return code %d' % (command, errno))


def execute(cmd, cwd, out, project_id, build_id):
    from .. import server

    log.debug('Forking subprocess [%s]' % cmd)
    subproc = Subprocess(
        cmd,
        stdout=Subprocess.STREAM,
        stderr=STDOUT,
        start_new_session=True,
        cwd=cwd)

    def end(data):
        if data == b'':
            log.debug('Got end')
            end.happened = True
        else:
            log.warn('Got WTF on ending: %s' % data)

    def send(data):
        out(data.decode('utf-8'))

    end.happened = False
    log.debug('Read until close [%s]' % cmd)
    subproc.stdout.read_until_close(end, send)

    log.debug('Polling [%s]' % cmd)
    while subproc.proc.poll() is None:
        sleep(.01)
        if (project_id, build_id) in server.builder.stop_flags:
            log.debug('Stopping [%s]' % cmd)
            os.killpg(subproc.proc.pid, SIGKILL)
            server.builder.stop_flags.remove((project_id, build_id))

    log.debug('Waiting for stream end [%s]' % cmd)
    while not end.happened:
        sleep(.01)

    log.debug('Ended with rv %d [%s]' % (subproc.proc.returncode, cmd))
    if subproc.proc.returncode != 0:
        raise ExecutionException(''.join(cmd), subproc.proc.returncode)
